/**************************************************************************

	nPOP

	Smtp.c (RFC 821, RFC 2821)

	Copyright (C) 1996-2002 by Tomoaki Nakashima. All rights reserved.
		http://www.nakka.com/
		nakka@nakka.com

**************************************************************************/

/**************************************************************************
	Include Files
**************************************************************************/

#include "General.h"
#include "global.h"
#include "md5.h"


/**************************************************************************
	Define
**************************************************************************/

#define SEND_BUFSIZE		4096				//M̕TCY

#define SMTP_ERRORSTATUS	400					//SMTPG[X|X

#define MIMEVER				"1.0"				//MIMEo[W

#define AUTH_CRAM_MD5		0					//SMTP-AUTH ̔F؃^Cv
#define AUTH_LOGIN			1
#define AUTH_PLAIN			2
#define AUTH_NON			99


/**************************************************************************
	Global Variables
**************************************************************************/

static struct TPMAILBOX *tpSendMailBox;			//M̃[{bNX
static struct TPMAILITEM *tpSendMailItem;		//M[
static BOOL SmtpAuthFlag;						//SMTP-AUTH
static TCHAR *body, *SendPoint;					//Mp{
static int SendLen;								//M
static int EndFlag;								//[M̃R}htO

//OQ
extern struct TPMAILBOX *MailBox;
extern int MailBoxCnt;
extern int SelBox;

extern SOCKET g_soc;
extern BOOL GetHostFlag;

extern int AutoSave;
extern int MailFlag;
extern TCHAR *g_Pass;

extern int IPCache;

extern int ESMTP;
extern TCHAR *SendHelo;
extern int SendMessageId;
extern int SendDate;


/**************************************************************************
	Local Function Prototypes
**************************************************************************/

static TCHAR *CreateAuthString(char *buf, TCHAR *User, TCHAR *Pass, TCHAR *ErrStr);
static TCHAR *CreatePlainString(TCHAR *User, TCHAR *Pass, TCHAR *ErrStr);
static BOOL GetAuthType(char *buf, int *type);
static TCHAR *GetPassword(struct TPMAILBOX *tpMailBox);
static int GetResponseStat(char *buf);
static BOOL SendHeader(SOCKET soc, TCHAR *Head, TCHAR *Sendbuf, TCHAR *ErrStr);
static BOOL SendConvHeader(SOCKET soc, TCHAR *Head, TCHAR *Sendbuf, BOOL Address, TCHAR *ErrStr);
static BOOL SendMailString(HWND hWnd, SOCKET soc, struct TPMAILITEM *tpMailItem, TCHAR *ErrStr);
static BOOL SendMailAddr(HWND hWnd, SOCKET soc, TCHAR *str, TCHAR *MailAddress, TCHAR *ErrStr);
static TCHAR *SendRcptTo(HWND hWnd, SOCKET soc, TCHAR *To, TCHAR *ErrStr);
static BOOL SendMailProc(HWND hWnd, SOCKET soc, char *buf, TCHAR *ErrStr,
							  struct TPMAILITEM *tpMailItem, BOOL ShowFlag);


/******************************************************************************

	HMAC_MD5

	MD5̃_CWFXg𐶐

******************************************************************************/

void HMAC_MD5(unsigned char *input, int len, unsigned char *key, int keylen, unsigned char *digest)
{
    MD5_CTX context;
    MD5_CTX tctx;
    unsigned char k_ipad[65];
    unsigned char k_opad[65];
    unsigned char tk[16];
    int i;

	//L[64oCg傫ꍇ̓L[̃_CWFXgL[ɂ
    if(keylen > 64){
        MD5Init(&tctx);
        MD5Update(&tctx, key, keylen);
        MD5Final(tk, &tctx);

        key = tk;
        keylen = 16;
    }

	//MD5(key XOR opad, MD5(key XOR ipad, input)) ̌vZs
    tFillMemory(k_ipad, sizeof(k_ipad), 0x36);
    tFillMemory(k_opad, sizeof(k_opad), 0x5c);
    for(i = 0; i < keylen; i++){
        *(k_ipad + i) ^= *(key + i);
        *(k_opad + i) ^= *(key + i);
    }

    MD5Init(&context);
    MD5Update(&context, k_ipad, 64);
    MD5Update(&context, input, len);
    MD5Final(digest, &context);

    MD5Init(&context);
    MD5Update(&context, k_opad, 64);
    MD5Update(&context, digest, 16);
    MD5Final(digest, &context);
}


/******************************************************************************

	CreateAuthString

	AUTH CRAM-MD5 ̃pX[h𐶐 (RFC 2554)

******************************************************************************/

static TCHAR *CreateAuthString(char *buf, TCHAR *User, TCHAR *Pass, TCHAR *ErrStr)
{
	TCHAR *ret;
	char *input, *key;
	char *p;
	unsigned char digest[16];
	int len;
	int i;

#ifdef UNICODE
	key = AllocTcharToChar(Pass);
	if(key == NULL){
		lstrcpy(ErrStr, STR_ERR_MEMALLOC);
		return NULL;
	}
#else
	key = Pass;
#endif

	for(p = buf; *p != '\0' && *p != ' '; p++);
	for(; *p == ' '; p++);

	input = (char *)LocalAlloc(LMEM_FIXED, tstrlen(p) + 1);
	if(input == NULL){
#ifdef UNICODE
		LocalFree(key);
#endif
		lstrcpy(ErrStr, STR_ERR_MEMALLOC);
		return NULL;
	}
	Base64Decode(p, input);

	//_CWFXgl̎擾
	HMAC_MD5(input, tstrlen(input), key, tstrlen(key), digest);
#ifdef UNICODE
	LocalFree(key);
#endif
	LocalFree(input);

	ret = (TCHAR *)LocalAlloc(LMEM_FIXED,
		sizeof(TCHAR) * (lstrlen(TEXT(" \r\n")) + lstrlen(User) + (16 * 2) + 1));
	if(ret == NULL){
		lstrcpy(ErrStr, STR_ERR_MEMALLOC);
		return NULL;
	}
	len = wsprintf(ret, TEXT("%s "), User);
	for(i = 0; i < 16; i++, len += 2){
		wsprintf(ret + len, TEXT("%02x"), digest[i]);
	}
	return ret;
}


/******************************************************************************

	CreatePlainString

	AUTH PLAIN ̕𐶐 (RFC 2554)

******************************************************************************/

static TCHAR *CreatePlainString(TCHAR *User, TCHAR *Pass, TCHAR *ErrStr)
{
#ifdef UNICODE
	TCHAR *ret;
#endif
	char *c_user, *c_pass;
	char *buf, *tmp;
	int len;

#ifdef UNICODE
	c_user = AllocTcharToChar(User);
	if(c_user == NULL){
		lstrcpy(ErrStr, STR_ERR_MEMALLOC);
		return NULL;
	}
	c_pass = AllocTcharToChar(Pass);
	if(c_pass == NULL){
		lstrcpy(ErrStr, STR_ERR_MEMALLOC);
		return NULL;
	}
#else
	c_user = User;
	c_pass = Pass;
#endif

	len = tstrlen(c_user) + tstrlen(c_pass) + 2;

	tmp = (char *)LocalAlloc(LPTR, len + 1);
	if(tmp == NULL){
#ifdef UNICODE
		LocalFree(c_user);
		LocalFree(c_pass);
#endif
		lstrcpy(ErrStr, STR_ERR_MEMALLOC);
		return NULL;
	}
	tstrcpy(tmp + 1, c_user);
	tstrcpy(tmp + 1 + tstrlen(c_user) + 1, c_pass);

#ifdef UNICODE
	LocalFree(c_user);
	LocalFree(c_pass);
#endif

	buf = (char *)LocalAlloc(LPTR, len * 2 + 4);
	if(buf == NULL){
		LocalFree(tmp);
		lstrcpy(ErrStr, STR_ERR_MEMALLOC);
		return NULL;
	}
	Base64Encode(tmp, buf, len);
	LocalFree(tmp);
#ifdef UNICODE
	ret = AllocCharToTchar(buf);
	LocalFree(buf);
	return ret;
#else
	return buf;
#endif
}


/******************************************************************************

	GetAuthType

	p\ SMTP AUTH ̎ނ擾 (RFC 2554)

******************************************************************************/

static BOOL GetAuthType(char *buf, int *type)
{
	char *p, *r;

	for(p = buf; *p != '\0' && *p != ' ' && *p != '-'; p++);
	for(; *p == ' ' || *p == '-'; p++);
	if(*p == '\0'){
		return FALSE;
	}
	if(StrCmpNI(p, "AUTH", 4) != 0){
		return FALSE;
	}
	for(; *p != '\0' && *p != ' ' && *p != '='; p++);
	for(; *p == ' ' || *p == '='; p++);
	if(*p == '\0'){
		return FALSE;
	}

	//CRAM-MD5
	r = p;
	while(*r != '\0'){
		if(StrCmpNI(r, "CRAM-MD5", 8) == 0){
			*type = AUTH_CRAM_MD5;
			break;
		}
		for(; *r != '\0' && *r != ' '; r++);
		for(; *r == ' '; r++);
	}
	if(*type == AUTH_CRAM_MD5){
		return TRUE;
	}

	//LOGIN
	r = p;
	while(*r != '\0'){
		if(StrCmpNI(r, "LOGIN", 5) == 0){
			*type = AUTH_LOGIN;
			break;
		}
		for(; *r != '\0' && *r != ' '; r++);
		for(; *r == ' '; r++);
	}
	if(*type == AUTH_LOGIN){
		return TRUE;
	}

	//PLAIN
	r = p;
	while(*r != '\0'){
		if(StrCmpNI(r, "PLAIN", 5) == 0){
			*type = AUTH_PLAIN;
			break;
		}
		for(; *r != '\0' && *r != ' '; r++);
		for(; *r == ' '; r++);
	}
	if(*type == AUTH_PLAIN){
		return TRUE;
	}
	return FALSE;
}


/******************************************************************************

	GetPassword

	pX[h̎擾

******************************************************************************/

static TCHAR *GetPassword(struct TPMAILBOX *tpMailBox)
{
	TCHAR *r;

	if(tpMailBox->AuthUserPass == 1){
		r = tpMailBox->SmtpPass;
		if(r == NULL || *r == TEXT('\0')){
			r = tpMailBox->SmtpTmpPass;
		}
	}else{
		r = tpMailBox->Pass;
		if(r == NULL || *r == TEXT('\0')){
			r = tpMailBox->TmpPass;
		}
	}
	if(r == NULL || *r == TEXT('\0')){
		r = g_Pass;
	}
	if(r == NULL || *r == TEXT('\0')){
		return NULL;
	}
	return r;
}


/******************************************************************************

	GetResponseStat

	SMTPX|X̉

******************************************************************************/

static int GetResponseStat(char *buf)
{
	char stat[5];
	int ret;

	//s̏ꍇ -1 Ԃ
	if(tstrlen(buf) > 3 && *(buf + 3) == '-'){
		return -1;
	}

	//Xe[^XR[h擾
	StrCpyN(stat, buf, 4);
	ret = a2i(stat);
	if(ret <= 0){
		ret = SMTP_ERRORSTATUS;
	}
	return ret;
}


/******************************************************************************

	SendHeader

	wb_ڂ̑M

******************************************************************************/

static BOOL SendHeader(SOCKET soc, TCHAR *Head, TCHAR *Sendbuf, TCHAR *ErrStr)
{
	TCHAR *buf;

	if(Sendbuf == NULL || *Sendbuf == TEXT('\0')){
		return TRUE;
	}

	buf = (TCHAR *)LocalAlloc(LMEM_FIXED, sizeof(TCHAR) * (lstrlen(Head) + 1 + lstrlen(Sendbuf) + 2 + 1));
	if(buf == NULL){
		lstrcpy(ErrStr, STR_ERR_MEMALLOC);
		return FALSE;
	}

	TStrJoin(buf, Head, TEXT(" "), Sendbuf, TEXT("\r\n"), (TCHAR *)-1);

	if(TSendBuf(soc, buf) == -1){
		LocalFree(buf);
		lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
		return FALSE;
	}
	LocalFree(buf);
	return TRUE;
}


/******************************************************************************

	SendConvHeader

	wb_ڂMIMEGR[hđM

******************************************************************************/

static BOOL SendConvHeader(SOCKET soc, TCHAR *Head, TCHAR *Sendbuf, BOOL Address, TCHAR *ErrStr)
{
	TCHAR *p;
	BOOL ret;

	if(Sendbuf == NULL || *Sendbuf == TEXT('\0')){
		return TRUE;
	}

	p = MIMEencode(Sendbuf, Address);
	if(p == NULL){
		lstrcpy(ErrStr, STR_ERR_MEMALLOC);
		return FALSE;
	}
	ret = SendHeader(soc, Head, p, ErrStr);
	LocalFree(p);
	return ret;
}


/******************************************************************************

	SendMailString

	wb_Ɩ{̑M (RFC 822, RFC 2822, RFC 2045, RFC 2076)

******************************************************************************/

static BOOL SendMailString(HWND hWnd, SOCKET soc, struct TPMAILITEM *tpMailItem, TCHAR *ErrStr)
{
	TCHAR buf[BUF_SIZE];
	TCHAR enc_buf[BUF_SIZE];
	TCHAR *p, *r;
	TCHAR *mctypr;
	TCHAR *mbody;
	int len;

	//From
	NULLCHECK_FREE(tpMailItem->From);
	tpMailItem->From = NULL;
	if(tpSendMailBox->MailAddress != NULL && *tpSendMailBox->MailAddress != TEXT('\0')){
		len = lstrlen(TEXT(" <>"));
		p = NULL;
		//[U̐ݒ
		if(tpSendMailBox->UserName != NULL){
			p = (TCHAR *)LocalAlloc(LMEM_FIXED, sizeof(TCHAR) * (lstrlen(tpSendMailBox->UserName) + 1));
			if(p != NULL){
				SetUserName(tpSendMailBox->UserName, p);
				len += lstrlen(p);
			}
		}
		len += lstrlen(tpSendMailBox->MailAddress);

		//From̍쐬ƑM
		tpMailItem->From = (TCHAR *)LocalAlloc(LMEM_FIXED, sizeof(TCHAR) * (len + 1));
		if(tpMailItem->From != NULL){
			r = tpMailItem->From;
			if(p != NULL && *p != TEXT('\0')){
				r = TStrJoin(r, p, TEXT(" "), (TCHAR *)-1);
			}

			TStrJoin(r, TEXT("<"), tpSendMailBox->MailAddress, TEXT(">"), (TCHAR *)-1);
			if(SendConvHeader(soc, TEXT(HEAD_FROM), tpMailItem->From, TRUE, ErrStr) == FALSE){
				return FALSE;
			}
		}
		NULLCHECK_FREE(p);
	}

	//To
	if(SendConvHeader(soc, TEXT(HEAD_TO), tpMailItem->To, TRUE, ErrStr) == FALSE){
		return FALSE;
	}

	//Cc
	if(SendConvHeader(soc, TEXT(HEAD_CC), tpMailItem->Cc, TRUE, ErrStr) == FALSE){
		return FALSE;
	}

	//Date
	GetTimeString(buf);
	if(SendDate == 1 && SendHeader(soc, TEXT(HEAD_DATE), buf, ErrStr) == FALSE){
		return FALSE;
	}
	NULLCHECK_FREE(tpMailItem->Date);
#ifdef UNICODE
	{
		char *cbuf, *dbuf;

		cbuf = AllocTcharToChar(buf);
		if(cbuf != NULL){
			dbuf = (char *)LocalAlloc(LMEM_FIXED, BUF_SIZE);
			if(dbuf != NULL){
				DateConv(cbuf, dbuf);
				tpMailItem->Date = AllocCharToTchar(dbuf);
				LocalFree(dbuf);
			}
			LocalFree(cbuf);
		}
	}
#else
	tpMailItem->Date = (char *)LocalAlloc(LMEM_FIXED, BUF_SIZE);
	if(tpMailItem->Date != NULL){
		DateConv(buf, tpMailItem->Date);
	}
#endif

	//Subject
	if(SendConvHeader(soc, TEXT(HEAD_SUBJECT), tpMailItem->Subject, FALSE, ErrStr) == FALSE){
		return FALSE;
	}

	//Reply-To
	if(tpMailItem->ReplyTo != NULL && *tpMailItem->ReplyTo != TEXT('\0')){
		//[ɐݒ肳Ă Reply-To
		if(SendConvHeader(soc, TEXT(HEAD_REPLYTO), tpMailItem->ReplyTo, TRUE, ErrStr) == FALSE){
			return FALSE;
		}
	}else{
		//[{bNXɐݒ肳Ă Reply-To
		if(SendConvHeader(soc, TEXT(HEAD_REPLYTO), tpSendMailBox->ReplyTo, TRUE, ErrStr) == FALSE){
			return FALSE;
		}
	}

	//In-Reply-To
	if(SendHeader(soc, TEXT(HEAD_INREPLYTO), tpMailItem->InReplyTo, ErrStr) == FALSE){
		return FALSE;
	}
	//References
	if(SendHeader(soc, TEXT(HEAD_REFERENCES), tpMailItem->References, ErrStr) == FALSE){
		return FALSE;
	}

	if(SendMessageId == 1){
		//Message-Id
		NULLCHECK_FREE(tpMailItem->MessageID);
		tpMailItem->MessageID = CreateMessageId((long)tpMailItem, tpSendMailBox->MailAddress);
		if(tpMailItem->MessageID == NULL){
			lstrcpy(ErrStr, STR_ERR_MEMALLOC);
			return FALSE;
		}
		if(SendHeader(soc, TEXT(HEAD_MESSAGEID), tpMailItem->MessageID, ErrStr) == FALSE){
			return FALSE;
		}
	}

	//MIME-Version
	if(SendHeader(soc, TEXT(HEAD_MIMEVERSION), TEXT(MIMEVER), ErrStr) == FALSE){
		return FALSE;
	}

	//{̃GR[h
	body = BodyEncode(tpMailItem->Body, buf, enc_buf, ErrStr);
	if(body == NULL){
		return FALSE;
	}

	//Ytt@Cꍇ̓}`p[gőM
	switch(CreateMultipart(tpMailItem->Attach, buf, enc_buf, &mctypr, body, &mbody))
	{
	case MP_ERROR_FILE:
		NULLCHECK_FREE(body);
		body = NULL;
		lstrcpy(ErrStr, STR_ERR_SOCK_NOATTACH);
		return FALSE;

	case MP_ERROR_ALLOC:
		NULLCHECK_FREE(body);
		body = NULL;
		lstrcpy(ErrStr, STR_ERR_MEMALLOC);
		return FALSE;

	case MP_NO_ATTACH:
		//Content-Type
		if(SendHeader(soc, TEXT(HEAD_CONTENTTYPE), buf, ErrStr) == FALSE){
			NULLCHECK_FREE(body);
			body = NULL;
			return FALSE;
		}
		//Content-Transfer-Encoding
		if(SendHeader(soc, TEXT(HEAD_ENCODING), enc_buf, ErrStr) == FALSE){
			NULLCHECK_FREE(body);
			body = NULL;
			return FALSE;
		}
		break;

	case MP_ATTACH:
		NULLCHECK_FREE(body);
		body = mbody;
		//Content-Type
		if(SendHeader(soc, TEXT(HEAD_CONTENTTYPE), mctypr, ErrStr) == FALSE){
			NULLCHECK_FREE(body);
			body = NULL;
			NULLCHECK_FREE(mctypr);
			return FALSE;
		}
		NULLCHECK_FREE(mctypr);
		break;
	}

	//X-Mailer (nPOP-Q)
	if(SendHeader(soc, TEXT(HEAD_X_MAILER), APP_NAMEEX, ErrStr) == FALSE){
		NULLCHECK_FREE(body);
		body = NULL;
		return FALSE;
	}

	if(TSendBuf(soc, TEXT("\r\n")) == -1){
		NULLCHECK_FREE(body);
		body = NULL;
		lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
		return FALSE;
	}

	SendPoint = body;
	SendLen = 0;

#ifdef WSAASYNC
	//M̃Cxg쐬
	if(WSAAsyncSelect(soc, hWnd, WM_SOCK_SELECT, FD_CONNECT | FD_READ | FD_WRITE | FD_CLOSE) == SOCKET_ERROR){
		NULLCHECK_FREE(body);
		body = NULL;
		lstrcpy(ErrStr, STR_ERR_SOCK_EVENT);
		return FALSE;
	}
#else
	//{M
	if(body != NULL && TSendBuf(soc, body) == -1){
		NULLCHECK_FREE(body);
		body = NULL;
		lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
		return FALSE;
	}
	NULLCHECK_FREE(body);
	body = NULL;

	if(TSendBuf(soc, TEXT("\r\n.\r\n")) == -1){
		lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
		return FALSE;
	}
#endif
	return TRUE;
}


/******************************************************************************

	SendMailAddr

	[AhX𑗐M

******************************************************************************/

static BOOL SendMailAddr(HWND hWnd, SOCKET soc, TCHAR *str, TCHAR *MailAddress, TCHAR *ErrStr)
{
	TCHAR *wbuf;

	wbuf = (TCHAR *)LocalAlloc(LMEM_FIXED,
		sizeof(TCHAR) * (lstrlen(TEXT(":<>\r\n")) + lstrlen(str) + lstrlen(MailAddress) + 1));
	if(wbuf == NULL){
		lstrcpy(ErrStr, STR_ERR_MEMALLOC);
		return FALSE;
	}
	// RCPT TO:<M惁[AhX>\r\n
	TStrJoin(wbuf, str, TEXT(":<"), MailAddress, TEXT(">\r\n"), (TCHAR *)-1);

	SetSocStatusTextT(hWnd, wbuf, 1);
	if(TSendBuf(soc, wbuf) == -1){
		LocalFree(wbuf);
		lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
		return FALSE;
	}
	LocalFree(wbuf);
	return TRUE;
}


/******************************************************************************

	SendRcptTo

	RCPT TO 𑗐M

******************************************************************************/

static TCHAR *SendRcptTo(HWND hWnd, SOCKET soc, TCHAR *To, TCHAR *ErrStr)
{
	TCHAR *p;

	//[AhX̒o
	p = (TCHAR *)LocalAlloc(LMEM_FIXED, sizeof(TCHAR) * (lstrlen(To) + 1));
	if(p == NULL){
		lstrcpy(ErrStr, STR_ERR_MEMALLOC);
		return (TCHAR *)-1;
	}
	To = GetMailAddress(To, p, FALSE);

	//[AhX̑M
	if(SendMailAddr(hWnd, soc, TEXT("RCPT TO"), p, ErrStr) == FALSE){
		LocalFree(p);
		return (TCHAR *)-1;
	}
	LocalFree(p);
	return ((*To != TEXT('\0')) ? To + 1 : To);
}


/******************************************************************************

	SendMailProc

	SMTP̉vV[W (RFC 821, RFC 2821)

******************************************************************************/

static BOOL SendMailProc(HWND hWnd, SOCKET soc, char *buf, TCHAR *ErrStr,
							  struct TPMAILITEM *tpMailItem, BOOL ShowFlag)
{
	HWND hListView;
	TCHAR *wbuf = NULL;
	TCHAR *user;
	TCHAR *pass;
	TCHAR *p, *r;
	static TCHAR *To, *Cc, *Bcc, *MyBcc;
	static int AuthType;
	int status = 0;
	int i;

	if(buf != NULL){
		SetStatusText(hWnd, buf);
		if(SmtpAuthFlag == TRUE && MailFlag == SMTP_EHLO){
			GetAuthType(buf, &AuthType);
		}
		if((status = GetResponseStat(buf)) == -1){
			return TRUE;
		}
	}
	switch(MailFlag){
	case SMTP_START:
		if(status >= SMTP_ERRORSTATUS){
			lstrcpy(ErrStr, STR_ERR_SOCK_RESPONSE);
			StrCatN(ErrStr, buf, BUF_SIZE - 1);
			return FALSE;
		}

		//[AhX̃`FbN
		if(tpSendMailBox->MailAddress == NULL || *tpSendMailBox->MailAddress == TEXT('\0')){
			lstrcpy(ErrStr, STR_ERR_SOCK_BADFROM);
			return FALSE;
		}

		//hC̒o
		if(SendHelo == NULL || *SendHelo == TEXT('\0')){
			p = tpSendMailBox->MailAddress;

			for(; *p != TEXT('@') && *p != TEXT('\0'); p++);
			if(*p == TEXT('\0')){
				lstrcpy(ErrStr, STR_ERR_SOCK_BADFROM);
				return FALSE;
			}
			p++;
		}else{
			p = SendHelo;
		}

		if(SmtpAuthFlag == FALSE && tpSendMailBox->SmtpAuth == 1){
			SmtpAuthFlag = TRUE;
			AuthType = AUTH_NON;
			r = TEXT("EHLO ");
			MailFlag = SMTP_EHLO;
		}else{
			SmtpAuthFlag = FALSE;
			r = (ESMTP != 0) ? TEXT("EHLO ") : TEXT("HELO ");
			MailFlag = SMTP_HELO;
		}
		wbuf = (TCHAR *)LocalAlloc(LMEM_FIXED, sizeof(TCHAR) * (lstrlen(r) + lstrlen(p) + lstrlen(TEXT("\r\n")) + 1));
		if(wbuf == NULL){
			lstrcpy(ErrStr, STR_ERR_MEMALLOC);
			return FALSE;
		}

		// HELO hC\r\n
		TStrJoin(wbuf, r, p, TEXT("\r\n"), (TCHAR *)-1);
		SetSocStatusTextT(hWnd, wbuf, 1);
		if(TSendBuf(soc, wbuf) == -1){
			LocalFree(wbuf);
			lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
			return FALSE;
		}
		LocalFree(wbuf);
		break;

	case SMTP_EHLO:
		if(status >= SMTP_ERRORSTATUS){
			//EHLO Ɏsꍇ HELO 𑗐M
			MailFlag = SMTP_START;
			return SendMailProc(hWnd, soc, NULL, ErrStr, tpMailItem, ShowFlag);
		}
		
		//IɔFؕݒ
		if(tpSendMailBox->SmtpAuthType != 0){
			AuthType = tpSendMailBox->SmtpAuthType - 1;
		}

		switch(AuthType)
		{
		case AUTH_CRAM_MD5:
		case AUTH_NON:
		default:
			//AUTH CRAM-MD5
			SetSocStatusTextT(hWnd, TEXT("AUTH CRAM-MD5"), 1);
			if(TSendBuf(soc, TEXT("AUTH CRAM-MD5\r\n")) == -1){
				lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
				return FALSE;
			}
			MailFlag = SMTP_AUTH;
			break;

		case AUTH_LOGIN:
			//AUTH LOGIN
			user = (tpSendMailBox->AuthUserPass == 1) ? tpSendMailBox->SmtpUser : tpSendMailBox->User;
			p = (TCHAR *)LocalAlloc(LPTR, sizeof(TCHAR) * (lstrlen(user) * 2 + 4));
			if(p == NULL){
				lstrcpy(ErrStr, STR_ERR_MEMALLOC);
				return FALSE;
			}
			TBase64Encode(user, p, 0);

			wbuf = (TCHAR *)LocalAlloc(LMEM_FIXED, sizeof(TCHAR) * (lstrlen(TEXT("AUTH LOGIN \r\n")) + lstrlen(p) + 1));
			if(wbuf == NULL){
				LocalFree(p);
				lstrcpy(ErrStr, STR_ERR_MEMALLOC);
				return FALSE;
			}
			//AUTH LOGIN User[CRLF] 𑗐M
			TStrJoin(wbuf, TEXT("AUTH LOGIN "), p, TEXT("\r\n"), (TCHAR *)-1);
			SetSocStatusTextT(hWnd, wbuf, 1);
			if(TSendBuf(soc, wbuf) == -1){
				LocalFree(p);
				LocalFree(wbuf);
				lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
				return FALSE;
			}
			LocalFree(p);
			LocalFree(wbuf);
			MailFlag = SMTP_AUTHLOGIN;
			break;

		case AUTH_PLAIN:
			//AUTH PLAIN
			user = (tpSendMailBox->AuthUserPass == 1) ? tpSendMailBox->SmtpUser : tpSendMailBox->User;
			pass = GetPassword(tpSendMailBox);
			if(pass == NULL || *pass == TEXT('\0')){
				lstrcpy(ErrStr, STR_ERR_SOCK_NOPASSWORD);
				return FALSE;
			}

			p = CreatePlainString(user, pass, ErrStr);
			if(p == NULL){
				return FALSE;
			}

			wbuf = (TCHAR *)LocalAlloc(LMEM_FIXED, sizeof(TCHAR) * (lstrlen(TEXT("AUTH PLAIN \r\n")) + lstrlen(p) + 1));
			if(wbuf == NULL){
				LocalFree(p);
				lstrcpy(ErrStr, STR_ERR_MEMALLOC);
				return FALSE;
			}

			//AUTH PLAIN Base64(<NULL>User<NULL>Pass)[CRLF] 𑗐M
			TStrJoin(wbuf, TEXT("AUTH PLAIN "), p, TEXT("\r\n"), (TCHAR *)-1);
			SetSocStatusTextT(hWnd, TEXT("AUTH PLAIN"), 1);
			if(TSendBuf(soc, wbuf) == -1){
				LocalFree(p);
				LocalFree(wbuf);
				lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
				return FALSE;
			}
			LocalFree(p);
			LocalFree(wbuf);
			MailFlag = SMTP_HELO;
			break;
		}
		break;

	case SMTP_AUTH:
		if(status >= SMTP_ERRORSTATUS){
			SmtpAuthFlag = FALSE;
			lstrcpy(ErrStr, STR_ERR_SOCK_SMTPAUTH);
			StrCatN(ErrStr, buf, BUF_SIZE - 1);
			return FALSE;
		}

		//AUTH CRAM-MD5̃pX[h쐬
		user = (tpSendMailBox->AuthUserPass == 1) ? tpSendMailBox->SmtpUser : tpSendMailBox->User;
		pass = GetPassword(tpSendMailBox);
		if(pass == NULL || *pass == TEXT('\0')){
			lstrcpy(ErrStr, STR_ERR_SOCK_NOPASSWORD);
			return FALSE;
		}
		wbuf = CreateAuthString(buf, user, pass, ErrStr);
		if(wbuf == NULL){
			return FALSE;
		}

	case SMTP_AUTHLOGIN:
		if(status >= SMTP_ERRORSTATUS){
			SmtpAuthFlag = FALSE;
			lstrcpy(ErrStr, STR_ERR_SOCK_SMTPAUTH);
			StrCatN(ErrStr, buf, BUF_SIZE - 1);
			return FALSE;
		}

		if(wbuf != NULL){
			r = wbuf;
		}else{
			r = GetPassword(tpSendMailBox);
			if(r == NULL || *r == TEXT('\0')){
				lstrcpy(ErrStr, STR_ERR_SOCK_NOPASSWORD);
				return FALSE;
			}
		}
		p = (TCHAR *)LocalAlloc(LPTR, sizeof(TCHAR) * (lstrlen(r) * 2 + 4));
		if(p == NULL){
			NULLCHECK_FREE(wbuf);
			lstrcpy(ErrStr, STR_ERR_MEMALLOC);
			return FALSE;
		}
		//Base64GR[h
		TBase64Encode(r, p, 0);
		NULLCHECK_FREE(wbuf);

		wbuf = (TCHAR *)LocalAlloc(LMEM_FIXED, sizeof(TCHAR) * (lstrlen(TEXT("\r\n")) + lstrlen(p) + 1));
		if(wbuf == NULL){
			LocalFree(p);
			lstrcpy(ErrStr, STR_ERR_MEMALLOC);
			return FALSE;
		}
		TStrJoin(wbuf, p, TEXT("\r\n"), (TCHAR *)-1);
		SetSocStatusTextT(hWnd, TEXT("Send password"), 1);
		if(TSendBuf(soc, wbuf) == -1){
			LocalFree(p);
			LocalFree(wbuf);
			lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
			return FALSE;
		}
		LocalFree(p);
		LocalFree(wbuf);
		MailFlag = SMTP_HELO;
		break;

	case SMTP_HELO:
		if(status >= SMTP_ERRORSTATUS){
			if(SmtpAuthFlag == FALSE){
				lstrcpy(ErrStr, STR_ERR_SOCK_HELO);
			}else{
				SmtpAuthFlag = FALSE;
				lstrcpy(ErrStr, STR_ERR_SOCK_SMTPAUTH);
			}
			StrCatN(ErrStr, buf, BUF_SIZE - 1);
			return FALSE;
		}

		//
		SetSocStatusTextT(hWnd, TEXT("RSET\r\n"), 1);
		if(TSendBuf(soc, TEXT("RSET\r\n")) == -1){
			lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
			return FALSE;
		}
		MailFlag = SMTP_RSET;
		break;

	case SMTP_RSET:
		if(status >= SMTP_ERRORSTATUS){
			lstrcpy(ErrStr, STR_ERR_SOCK_RSET);
			StrCatN(ErrStr, buf, BUF_SIZE - 1);
			return FALSE;
		}

		//ԐM惁[AhX̑M
		if(SendMailAddr(hWnd, soc, TEXT("MAIL FROM"), tpSendMailBox->MailAddress, ErrStr) == FALSE){
			return FALSE;
		}
		MailFlag = SMTP_MAILFROM;
		break;

	case SMTP_MAILFROM:
		if(status >= SMTP_ERRORSTATUS){
			lstrcpy(ErrStr, STR_ERR_SOCK_MAILFROM);
			StrCatN(ErrStr, buf, BUF_SIZE - 1);
			return FALSE;
		}

		if(tpMailItem->To == NULL){
			lstrcpy(ErrStr, STR_ERR_SOCK_NOTO);
			return FALSE;
		}
		//ݒ
		To = tpMailItem->To;
		Cc = tpMailItem->Cc;
		Bcc = tpMailItem->Bcc;
		//ĂɑM
		MyBcc = NULL;
		if(tpSendMailBox->MyAddr2Bcc == 1){
			MyBcc = (tpSendMailBox->BccAddr != NULL && *tpSendMailBox->BccAddr != TEXT('\0')) ?
				tpSendMailBox->BccAddr : tpSendMailBox->MailAddress;
		}
		MailFlag = SMTP_RCPTTO;
		return SendMailProc(hWnd, soc, NULL, ErrStr, tpMailItem, ShowFlag);

	case SMTP_RCPTTO:
		if(status >= SMTP_ERRORSTATUS){
			lstrcpy(ErrStr, STR_ERR_SOCK_RCPTTO);
			StrCatN(ErrStr, buf, BUF_SIZE - 1);
			return FALSE;
		}

		//To Ɏw肳ꂽ[AhX̑M
		if(To != NULL && *To != TEXT('\0')){
			if((To = SendRcptTo(hWnd, soc, To, ErrStr)) == (TCHAR *)-1){
				return FALSE;
			}
			MailFlag = SMTP_RCPTTO;
			break;
		}

		//Cc Ɏw肳ꂽ[AhX̑M
		if(Cc != NULL && *Cc != TEXT('\0')){
			if((Cc = SendRcptTo(hWnd, soc, Cc, ErrStr)) == (TCHAR *)-1){
				return FALSE;
			}
			MailFlag = SMTP_RCPTTO;
			break;
		}

		//Bcc Ɏw肳ꂽ[AhX̑M
		if(Bcc != NULL && *Bcc != TEXT('\0')){
			if((Bcc = SendRcptTo(hWnd, soc, Bcc, ErrStr)) == (TCHAR *)-1){
				return FALSE;
			}
			MailFlag = SMTP_RCPTTO;
			break;
		}

		//
		if(MyBcc != NULL && *MyBcc != TEXT('\0')){
			if((MyBcc = SendRcptTo(hWnd, soc, MyBcc, ErrStr)) == (TCHAR *)-1){
				return FALSE;
			}
			MailFlag = SMTP_RCPTTO;
			break;
		}

		//DATAM
		SetSocStatusTextT(hWnd, TEXT("DATA\r\n"), 1);
		if(TSendBuf(soc, TEXT("DATA\r\n")) == -1){
			lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
			return FALSE;
		}
		MailFlag = SMTP_DATA;
		break;

	case SMTP_DATA:
		if(status >= SMTP_ERRORSTATUS){
			lstrcpy(ErrStr, STR_ERR_SOCK_DATA);
			StrCatN(ErrStr, buf, BUF_SIZE - 1);
			return FALSE;
		}

		SetSocStatusTextT(hWnd, STR_STATUS_SENDBODY, 1);
		SwitchCursor(FALSE);
		if(SendMailString(hWnd, soc, tpMailItem, ErrStr) == FALSE){
			SwitchCursor(TRUE);
			return FALSE;
		}
		SwitchCursor(TRUE);
#ifdef WSAASYNC
		MailFlag = SMTP_SENDBODY;
		break;

	case SMTP_SENDBODY:
		break;
#else
		MailFlag = EndFlag;
		break;
#endif

	case SMTP_NEXTSEND:
		if(status >= SMTP_ERRORSTATUS){
			lstrcpy(ErrStr, STR_ERR_SOCK_MAILSEND);
			StrCatN(ErrStr, buf, BUF_SIZE - 1);
			return FALSE;
		}

		tpMailItem->Status = tpMailItem->MailStatus = ICON_SENDMAIL;
		if(ShowFlag == TRUE){
			hListView = GetDlgItem(hWnd, IDC_LISTVIEW);
			i = ListView_GetMemToItem(hListView, tpMailItem);
			if(i != -1){
				ListView_RedrawItems(hListView, i, i);
				UpdateWindow(hListView);
			}
		}
		if(AutoSave == 1){
			//Mt@Cɕۑ
			SaveMail(SENDBOX_FILE, MailBox + MAILBOX_SEND, 2);
		}

		//̑M[̎擾
		i = Item_GetNextMailBoxSendItem((MailBox + MAILBOX_SEND), -1, GetNameToMailBox(tpSendMailBox->Name));
		if(i == -1){
			//[MI
			SetSocStatusTextT(hWnd, TEXT(QUIT), 1);
			if(TSendBuf(soc, TEXT(QUIT)) == -1){
				lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
				return FALSE;
			}
			MailFlag = SMTP_QUIT;
			break;
		}
		if(tpSendMailItem == *((MailBox + MAILBOX_SEND)->tpMailItem + i)){
			lstrcpy(ErrStr, STR_ERR_SOCK_GETITEMINFO);
			return FALSE;
		}
		tpSendMailItem = *((MailBox + MAILBOX_SEND)->tpMailItem + i);
		if((tpSendMailItem->To == NULL || *tpSendMailItem->To == TEXT('\0')) &&
			(tpSendMailItem->Cc == NULL || *tpSendMailItem->Cc == TEXT('\0')) &&
			(tpSendMailItem->Bcc == NULL || *tpSendMailItem->Bcc == TEXT('\0'))){
			lstrcpy(ErrStr, STR_ERR_SOCK_NOTO);
			return FALSE;
		}
		//
		SetSocStatusTextT(hWnd, TEXT("RSET\r\n"), 1);
		if(TSendBuf(soc, TEXT("RSET\r\n")) == -1){
			lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
			return FALSE;
		}
		MailFlag = SMTP_RSET;
		break;

	case SMTP_SENDEND:
		if(status >= SMTP_ERRORSTATUS){
			lstrcpy(ErrStr, STR_ERR_SOCK_MAILSEND);
			StrCatN(ErrStr, buf, BUF_SIZE - 1);
			return FALSE;
		}
		SetSocStatusTextT(hWnd, TEXT(QUIT), 1);
		if(TSendBuf(soc, TEXT(QUIT)) == -1){
			lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
			return FALSE;
		}
		tpMailItem->Status = tpMailItem->MailStatus = ICON_SENDMAIL;
		if(ShowFlag == TRUE){
			hListView = GetDlgItem(hWnd, IDC_LISTVIEW);
			i = ListView_GetMemToItem(hListView, tpMailItem);
			if(i != -1){
				ListView_RedrawItems(hListView, i, i);
				UpdateWindow(hListView);
			}
		}
		if(AutoSave == 1){
			//Mt@Cɕۑ
			SaveMail(SENDBOX_FILE, MailBox + MAILBOX_SEND, 2);
		}
		MailFlag = SMTP_QUIT;
		break;

	case SMTP_QUIT:
		break;
	}
	return TRUE;
}


/******************************************************************************

	SmtpError

	M[ɃG[Xe[^Xt

******************************************************************************/

void SmtpError(HWND hWnd)
{
	HWND hListView;
	int i;

	NULLCHECK_FREE(body);
	body = NULL;
	SmtpAuthFlag = FALSE;

	if(tpSendMailItem == NULL){
		return;
	}
	tpSendMailItem->Status = tpSendMailItem->MailStatus = ICON_ERROR;

	if(SelBox == MAILBOX_SEND){
		hListView = GetDlgItem(hWnd, IDC_LISTVIEW);
		i = ListView_GetMemToItem(hListView, tpSendMailItem);
		if(i != -1){
			ListView_RedrawItems(hListView, i, i);
			UpdateWindow(hListView);
		}
	}
	tpSendMailItem = NULL;
}


/******************************************************************************

	SmtpProc

	[MvV[W

******************************************************************************/

BOOL SmtpProc(HWND hWnd, SOCKET soc, char *buf, int len, TCHAR *ErrStr,
							  struct TPMAILBOX *tpMailBox, BOOL ShowFlag)
{
	BOOL ret;

	if(tpSendMailItem == NULL){
		return FALSE;
	}
	ret = SendMailProc(hWnd, soc, buf, ErrStr, tpSendMailItem, ShowFlag);
	if(ret == FALSE){
		SendBuf(soc, QUIT);
		SmtpError(hWnd);
	}
	return ret;
}


/******************************************************************************

	SendMailItem

	P̑M[̏

******************************************************************************/

SOCKET SendMailItem(HWND hWnd, struct TPMAILBOX *tpMailBox, struct TPMAILITEM *tpMailItem, int EndMailFlag, TCHAR *ErrStr)
{
	SOCKET soc;

	tpSendMailBox = tpMailBox;
	tpSendMailItem = tpMailItem;
	SmtpAuthFlag = FALSE;
	EndFlag = EndMailFlag;

	if((tpSendMailItem->To == NULL || *tpSendMailItem->To == TEXT('\0')) &&
		(tpSendMailItem->Cc == NULL || *tpSendMailItem->Cc == TEXT('\0')) &&
		(tpSendMailItem->Bcc == NULL || *tpSendMailItem->Bcc == TEXT('\0'))){
		lstrcpy(ErrStr, STR_ERR_SOCK_NOTO);
		return -1;
	}

	//zXgIPAhX擾 (擾IP͕ۑĂ)
	if(tpSendMailBox->SmtpIP == 0 || IPCache == 0){
		GetHostFlag = TRUE;
		tpSendMailBox->SmtpIP = GetHostByName(hWnd, tpSendMailBox->SmtpServer, ErrStr);
		GetHostFlag = FALSE;
		if(tpSendMailBox->SmtpIP == 0){
			return -1;
		}
	}
	//ڑJn
	soc = ConnectServer(hWnd,
		tpSendMailBox->SmtpIP, (unsigned short)tpSendMailBox->SmtpPort,
		ErrStr);
	if(soc == -1){
		return -1;
	}
	return soc;
}

/******************************************************************************

	SendProc

	MvV[W

******************************************************************************/

#ifdef WSAASYNC
BOOL SendProc(HWND hWnd, SOCKET soc, TCHAR *ErrStr, struct TPMAILBOX *tpMailBox)
{
	TCHAR buf[SEND_BUFSIZE];
	int len;

	if(body == NULL && SendPoint != NULL){
		lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
		return FALSE;
	}

	//MI
	if(body == NULL || *SendPoint == TEXT('\0')){
		if(Tsend(soc, TEXT("\r\n.\r\n")) == -1){
			if(WSAGetLastError() == WSAEWOULDBLOCK){
				return TRUE;
			}
			lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
			return FALSE;
		}
		LocalFree(body);
		body = NULL;
		MailFlag = EndFlag;
		if(WSAAsyncSelect(soc, hWnd, WM_SOCK_SELECT, FD_CONNECT | FD_READ | FD_CLOSE) == SOCKET_ERROR){
			LocalFree(body);
			body = NULL;
			lstrcpy(ErrStr, STR_ERR_SOCK_EVENT);
			return FALSE;
		}
		return TRUE;
	}

	//đM
	TStrCpyN(buf, SendPoint, SEND_BUFSIZE - 1);
	if((len = Tsend(soc, buf)) == -1){
		if(WSAGetLastError() == WSAEWOULDBLOCK){
			return TRUE;
		}
		LocalFree(body);
		body = NULL;
		lstrcpy(ErrStr, STR_ERR_SOCK_SEND);
		return FALSE;
	}

	SendPoint += len;
	SendLen += len;
	SetStatusRecvLen(hWnd, SendLen, STR_STATUS_SOCKINFO_SEND);
	if(WSAAsyncSelect(soc, hWnd, WM_SOCK_SELECT, FD_CONNECT | FD_READ | FD_WRITE | FD_CLOSE) == SOCKET_ERROR){
		LocalFree(body);
		body = NULL;
		lstrcpy(ErrStr, STR_ERR_SOCK_EVENT);
		return FALSE;
	}
	return TRUE;
}
#endif
/* End of source */
